;Very small composite video signal generator
;pixel graphics up to 80 pixels wide
;ATTiny11(L) overclocked to 8MHz (external XTAL), Fuse Byte = 0x1F
;use AVRA (http://avra.sourceforge.net/) for assembly
;          ___
;  PB1 ---|___|---o-----> Composite Output (load with 75 ohms)
;          1k     |
;          ___    |
;  PB0 ---|___|---'
;          390
;
;Possible graphics resolutions:
;
;   Width  |     Height
;  (fixed) |  max  | 4:3 ratio
; ---------+-------+-----------
;       8  |  288  |       6
;      16  |  288  |      12
;      24  |  242  |      18
;      32  |  183  |      24
;      40  |  140  |      30
;      48  |  119  |      36
;      56  |  104  |      42
;      64  |   91  |      48
;      72  |   83  |      54
;      80  |   75  |      60
;
;(max heights were calculated from available flash memory)
;
;Version History:
;  2008-??-??  first version
;  2008-06-03  added support for variable image width, optimised code size

.include "../include/tn11def.inc"

;R0 needed for LPM
.def	zero = R1
.def	backup = R2

.def	temp = R16
.def	temp2 = R17
.def	count = R18
.def	lineL = R19
.def	lineH = R20
.def	linediv = R21
.def	one = R22
.def	linestartL = R23
.def	linestartH = R24
.def	pixelwait = R25

;R31:R30 needed for pointer (Z)

.equ	PixelWidth = 80
.equ	PixelHeight = 72
.equ	LineDivInit = 288/PixelHeight ;scanlines per image line (pixel height)
.equ	FirstLine = 18 + (288-PixelHeight*LineDivInit)/2 ;first used scanline
.equ	LastLine = FirstLine + PixelHeight*LineDivInit ;last used scanline
.equ	CyclesPerPixel = (8*50)/PixelWidth ;XTAL cycles per pixel (pixel width)
.equ	PrePixelCycles = ((8*50) - (PixelWidth*CyclesPerPixel))/2
;Explanations: * 288 is total number of usable scanlines per frame.
;              * (8*50) is XTAL cycles per scanline (8 MHz * 50 us),
;                the remaining 2 us in a 52 us line cannot be used.

.equ	DATA = 0 ;PORTB0 ;don't change
.equ	SYNC = 1 ;PORTB1

;===============================================================================

.if (LineDivInit == 0)
	.error "Image is too tall (max. 288 lines)"
.elseif (CyclesPerPixel < 5)
	.error "Image is too wide (max. 80 pixels)"
.elseif (PixelWidth % 8 != 0)
	.error "Image width is not a multiple of 8"
.endif
.message "Pixel Geometry: ", CyclesPerPixel, " cycles x ", LineDivInit, " lines"
.message "Image Size:     ", PixelWidth*CyclesPerPixel, " cycles x ", PixelHeight*LineDivInit, " lines"
.message "Image Position: +", PrePixelCycles, " cycles x +", FirstLine-18, " lines"

;macro for extra pixel delay if image is less than 80 pixels wide
.macro AddLength
.if ((CyclesPerPixel-5) == 1)
	nop
.elseif ((CyclesPerPixel-5) == 2)
	rjmp	PC+1
.elseif ((CyclesPerPixel-5) == 3)
	rjmp	PC+1
	nop
.elseif ((CyclesPerPixel-5) == 4)
	rjmp	PC+1
	rjmp	PC+1
.elseif ((CyclesPerPixel-5) == 5)
	rjmp	PC+1
	rjmp	PC+1
	nop
.elseif ((CyclesPerPixel-5) == 6)
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
.elseif ((CyclesPerPixel-5) >= 7)
	rcall	pixelwait_long
.endif
.endmacro

;===============================================================================

.cseg
.org 0x000
	clr	zero
	ldi	one, 1
	rjmp	reset
	
.org OVF0addr ;.org 0x003
	;check if vsync is needed
	add	lineL, one			;1
	adc	lineH, zero			;2
	cpi	lineL, LOW(311)			;3
	sbrc	lineH, 0 ;HIGH(311) = 1 => bit 0 set	;4 (5)
	brsh	vsync				;5 [6]

	;horizontal synchronisation
	lpm					;8 (8)
	out	PORTB, zero			;9 (9) ** START SYNC PULSE **
	;HSYNC pulse: 4.75us = 38 cycles
	;reload timer for HSYNC (less than 64 cycles after interrupt call)
	ldi	temp, 256-8			;1
	out	TCNT0, temp			;2
	ldi	temp, 11			;2 + 3*11 = 35
line_hsync:	
	dec	temp
	brne	line_hsync
	nop					;36
	sbi	PORTB, SYNC			;38
	
	;PRE black + 1us: 5.75us + 1us = 46 + 8 cycles = 54 cycles, additional
	;                 delay if image is less than 80 pixels wide
	;prepare flash pointer and byte counter
	mov	ZL, linestartL			; 1
	mov	ZH, linestartH			; 2
	ldi	count, PixelWidth/8		; 3
	;check image boundaries
	ldi	temp, HIGH(FirstLine)		; 4
	cpi	lineL, LOW(FirstLine)		; 5
	cpc	lineH, temp			; 6
	brlo	line_end			; 7
	ldi	temp, HIGH(LastLine)		; 8
	cpi	lineL, LOW(LastLine)		; 9
	cpc	lineH, temp			;10
	brsh	line_end			;11
	;load first data byte
	lpm					;14
	;determine next line
	dec	linediv				;15
	brne	line_noinc			;16 (17)
	ldi	temp, PixelWidth/8		;17
	ldi	linediv, LineDivInit		;18
	add	linestartL, temp		;19
	adc	linestartH, zero		;20
	rjmp	line_cont			;22
	
line_noinc:
	lpm					;   (20)
	rjmp	line_cont			;   (22)

line_end:
	reti

;--------------------
	
vsync:
	;vertical synchronisation
	;reload timer for VSYNC (less than 64 cycles after interrupt call)
	;(VSYNC pulse is the length of 8 lines)
	ldi	temp, 256-8*8			;[7]
	out	TCNT0, temp			;[8]
	out	PORTB, zero			;[9] ** START SYNC PULSE **
	;SYNC: 2.375us = 19 cycles
	ldi	temp, 4			;4*3 = 12 (vsync_prepost delay counter)
	rjmp	PC+1			;14
	;5 pre pulses (short sync)
	rcall	vsync_prepost_sync	;17
	;rcall	vsync_prepost_timer
	rcall	vsync_prepost
	rcall	vsync_prepost
	rcall	vsync_prepost
	rcall	vsync_prepost
	rcall	vsync_main ;5 main pulses (short black)
	rcall	vsync_main
	rcall	vsync_main
	rcall	vsync_main
	rcall	vsync_main
	rcall	vsync_prepost ;5 post pulses (short sync)
	rcall	vsync_prepost
	rcall	vsync_prepost
	rcall	vsync_prepost
	rcall	vsync_prepost
	;initialize registers
	ldi	lineL, 0
	ldi	lineH, 0
	ldi	linediv, LineDivInit
	ldi	linestartL, LOW(image*2)
	ldi	linestartH, HIGH(image*2)
	reti

;--------------------
	
line_cont:
	;continue scanline: output graphics data
	ldi	temp, 7 + PrePixelCycles/4	;22 + (7*4) = 50 (+ PrePixelCycles)
line_wait:
	nop
	dec	temp
	brne	line_wait
	
	ldi	temp2, 1<<SYNC		;Pixel 1 ;51 (+ PrePixelCycles)
line_loop:
	sbrc	R0, 7		;2		;52 (+ PrePixelCycles)
	ldi	temp2, 1<<SYNC | 1<<DATA ;3	;53 (+ PrePixelCycles)
	out	PORTB, temp2	;4		;54 (+ PrePixelCycles)
		; ** first pixel starts here **
	AddLength
	mov	temp, R0	;5
	ldi	temp2, 1<<SYNC		;Pixel 2
	sbrc	R0, 6		;2
	ldi	temp2, 1<<SYNC | 1<<DATA ;3
	out	PORTB, temp2	;4
	AddLength
	andi	temp, 0x01	;5
	ldi	temp2, 1<<SYNC		;Pixel 3
	sbrc	R0, 5		;2
	ldi	temp2, 1<<SYNC | 1<<DATA ;3
	out	PORTB, temp2	;4
	AddLength
	ori	temp, 1<<SYNC	;5
	ldi	temp2, 1<<SYNC		;Pixel 4
	sbrc	R0, 4		;2
	ldi	temp2, 1<<SYNC | 1<<DATA ;3
	out	PORTB, temp2	;4
	AddLength
	add	ZL, one		;5 ;the following 4 instructions don't change any flags
	ldi	temp2, 1<<SYNC		;Pixel 5
	sbrc	R0, 3		;2
	ldi	temp2, 1<<SYNC | 1<<DATA ;3
	out	PORTB, temp2	;4
	AddLength
	adc	ZH, zero	;5
	ldi	temp2, 1<<SYNC		;Pixel 6
	sbrc	R0, 2		;2
	ldi	temp2, 1<<SYNC | 1<<DATA ;3
	out	PORTB, temp2	;4
	AddLength
	dec	count		;5 ;the following 7 instructions don't change any flags
	ldi	temp2, 1<<SYNC		;Pixel 7
	sbrc	R0, 1		;2
	ldi	temp2, 1<<SYNC | 1<<DATA ;3
	out	PORTB, temp2	;4
	AddLength
	lpm			;5	;Pixel 8 (2 cycles)
	ldi	temp2, 1<<SYNC	;3
	out	PORTB, temp	;4
	AddLength
	brne	line_loop	;5	;Pixel 1
	
	lpm ;wait for last pixel to finish (3 cycles)
	out	PORTB, temp2 ;output black for the rest of the line

	reti
	
;-------------------------------------------------------------------------------

vsync_prepost:
	out	PORTB, zero
	;SYNC: 2.375us = 19 cycles
	rjmp	PC+1			;2
	ldi	temp, 5			;2 + 5*3 = 17
vsync_prepost_sync:
	dec	temp
	brne	vsync_prepost_sync
	sbi	PORTB, SYNC		;19
	;BLACK: 29.625us = 237 cycles
	ldi	temp, 76		;76*3 = 228
vsync_prepost_black:
	dec	temp
	brne	vsync_prepost_black
	nop				;229
	ret				;237 (incl. next 'rcall' and 'out')

;--------------------

vsync_main:
	out	PORTB, zero
	;SYNC: 27.25us = 218 cycles
	ldi	temp, 72		;72*3 = 216
vsync_sync:
	dec	temp
	brne	vsync_sync
	sbi	PORTB, SYNC		;218
	;BLACK: 4.75us = 38 cycles
	ldi	temp, 10		;10*3 = 30
vsync_black:
	dec	temp
	brne	vsync_black
	ret				;38 (incl. next 'rcall' and 'out')
	
;-------------------------------------------------------------------------------

pixelwait_long:
	;additional pixel delay is 7 or more cycles (rcall = 3, ret = 4)	
	;7 cycles: do nothing (rcall + ret is already 7 cycles)
	;8..11 cycles: "flat" instructions, no loop, no flags changed
.if ((CyclesPerPixel-5) == 8)
	nop
	ret
.elseif ((CyclesPerPixel-5) == 9)
	rjmp	PC+1
	ret
.elseif ((CyclesPerPixel-5) == 10)
	rjmp	PC+1
	nop
	ret
.elseif ((CyclesPerPixel-5) == 11)
	rjmp	PC+1
	rjmp	PC+1
	ret
.elseif ((CyclesPerPixel-5) >= 12)
	;12 or more cycles: delay loop, backup SREG to preserve flags
.if ((CyclesPerPixel-5-7-2) % 3 == 1)
	nop
.elseif ((CyclesPerPixel-5-7-2) % 3 == 2)
	rjmp	PC+1
.endif
	in	backup, SREG ;backup flags
	ldi	pixelwait, (CyclesPerPixel-5-7-2)/3
pixelwait_loop:
	dec	pixelwait
	brne	pixelwait_loop
	out	SREG, backup ;restore flags
	ret
.endif
	
;===============================================================================

reset:
	;init registers
;	ldi	lineL, 0
;	ldi	lineH, 0
;	ldi	charline, 7
;	ldi	linediv, LineDivInit
;	ldi	textpos, LOW(text*2)
	
	;init PORTB
	ldi	temp, 0x1F
	out	DDRB, temp
	;out	PORTB, zero
	
	;init timer
;	ldi	temp, 256-8 ;Clk/64/8 = 15625 Hz (horizontal line frequency)
;	out	TCNT0, temp
	ldi	temp, 0x03 ;Clk/64
	out	TCCR0, temp
;	ldi	temp, 1<<TOIE0
	out	TIMSK, temp
	
	;enable interrupts
	sei
	
loop:
	rjmp	loop
	
;===============================================================================
	
;.org 0x200-(PixelWidth/8 * PixelHeight)/2
image:
.include "80x72.png.asm"
